-- This file is part of Hypha.
--
-- Hypha is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Hypha is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Hypha.  If not, see <http://www.gnu.org/licenses/>.
--
-- Copyright 2013 Luca Padovani

{
module Parser where

import Lexer
import Process
}

%name process
%tokentype { Token }
%error { throwError }

%token 
  new       { NEW }
  intval    { Lexer.INT $$ }
  true      { TRUE }
  false     { FALSE }
  var       { ID $$ }
  let       { LET }
  in        { IN }
  inl       { INL }
  inr       { INR }
  fst       { FST }
  snd       { SND }
  case      { CASE }
  of        { OF }
  if        { IF }
  then      { THEN }
  else      { ELSE }
  arrow     { ARROW }
  '='       { Lexer.EQ }
  '|'       { PAR }
  '('       { LPAR }
  ')'       { RPAR }
  '<'       { Lexer.LT }
  '>'       { Lexer.GT }
  le        { Lexer.LE }
  ge        { Lexer.GE }
  ne        { Lexer.NE }
  and       { AND }
  or        { OR }
  not       { NOT }
  '['       { LBRACK }
  ']'       { RBRACK }
  '{'       { LBRACE }
  '}'       { RBRACE }
  '!'       { EMARK }
  '?'       { QMARK }
  ','       { COMMA }
  '.'       { DOT }
  ':'       { COLON }
  ';'       { SEMICOLON }
  '*'       { STAR }
  '+'       { PLUS }
  '-'       { MINUS }
  '/'       { SLASH }
  '_'       { UNDERSCORE }

%left '|'
%nonassoc in
%right ','
%nonassoc inl inr fst snd
%left and or
%nonassoc '<' '>' '=' le ge ne
%left '+' '-'
%left '*' '/'
%nonassoc not
%right BANG
%right '.'

%%

Process : Expr '!' Expr Continuation { Send $1 $3 $4 }
        | Expr '?' '(' PatternOpt ')' Continuation { expandReceive $1 $4 $6 }
        | let Pattern '=' Expr in Process { expandLet $2 $4 $6 }
        | case Expr of '[' Pattern arrow Process ';' Pattern arrow Process ']'
	{ let (x1, p1) = expandCase $5 $7 in
	  let (x2, p2) = expandCase $9 $11 in
	  Case $2 x1 () p1 x2 () p2 }
        | if Expr then Process else Process { Case $2 "_" () $4 "_" () $6 }
        | Process '|' Process { Par $1 $3 } 
        | new var in Process { New $2 () $4 }
        | '{' ProcessOpt '}' { Group $2 }
        | '*' Process %prec BANG { Star $2 }

Continuation : { Idle }
        | '.' Process { $2 }

ProcessOpt :  { Idle }
        | Process { $1 }
        
PatternOpt : { PVar "_" }
           | Pattern { $1 }

Pattern : '_' { PVar "_" }
        | var { PVar $1 }
        | Pattern ',' Pattern { PPair $1 $3 }
        | '(' Pattern ')' { $2 }

Expr    : '(' ')' { UNIT }
        | intval   { Process.INT $1 }
        | true { Process.Left UNIT }
        | false { Process.Right UNIT }
        | var   { Id $1 }
	| Expr and Expr { LogBinOp $1 $3 }
	| Expr or Expr { LogBinOp $1 $3 }
	| Expr '<' Expr { RelOp $1 $3 }
	| Expr '>' Expr { RelOp $1 $3 }
	| Expr '=' Expr { RelOp $1 $3 }
	| Expr le Expr { RelOp $1 $3 }
	| Expr ge Expr { RelOp $1 $3 }
	| Expr ne Expr { RelOp $1 $3 }
	| not Expr { LogUnOp $2 }
	| Expr '+' Expr { IntBinOp $1 $3 }
	| Expr '-' Expr { IntBinOp $1 $3 }
	| Expr '*' Expr { IntBinOp $1 $3 }
	| Expr '/' Expr { IntBinOp $1 $3 }
        | inl Expr { Process.Left $2 }
        | inr Expr { Process.Right $2 }
        | fst Expr { Process.First $2 }
        | snd Expr { Process.Second $2 }
        | '(' Expr ')' { $2 }
        | Expr ',' Expr { Pair $1 $3 }
        
{

data Pattern = PVar String
             | PPair Pattern Pattern

expandReceive :: Expression -> Pattern -> Process () -> Process ()
expandReceive u (PVar x) p = Receive u x () p
expandReceive u pattern p = Receive u "_msg" () $ expandLet pattern (Id "_msg") p

expandCase :: Pattern -> Process () -> (String, Process ())
expandCase (PVar x) p = (x, p)
expandCase pattern p = ("_var", expandLet pattern (Id "_var") p)

expandLet :: Pattern -> Expression -> Process () -> Process ()
expandLet = expand "_var"
  where
    expand _ (PVar x) e p = Let x () "_" () (Pair e UNIT) p
    expand _ (PPair (PVar x) (PVar y)) e p = Let x () y () e p
    expand path (PPair p1 p2) e p = let pathl = path ++ ".1" in
                                    let pathr = path ++ ".2" in
                                    Let pathl () pathr () e (expand pathl p1 (Id pathl)
							     (expand pathr p2 (Id pathr) p))

throwError :: [Token] -> a
throwError tokens = error ("Parse error on "++ (show $ head tokens) ++ "\n")

}
